package com.acme.wsapigatewaytomcat.websocket;

import java.io.IOException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.socket.CloseStatus;
import org.springframework.web.socket.WebSocketHandler;
import org.springframework.web.socket.WebSocketMessage;
import org.springframework.web.socket.WebSocketSession;
import org.springframework.web.socket.handler.WebSocketHandlerDecorator;

public class ClientWebSocketConnectionHandler extends WebSocketHandlerDecorator {

    private static final Logger LOGGER = LoggerFactory.getLogger( ClientWebSocketConnectionHandler.class );
    private static final String SERVER_WEBSOCKET_SESSION_KEY = "server-websocket-session";
    private static final String SERVER_WEBSOCKET_HANDLER_KEY = "server-websocket-handler";
    private static final String WEBSOCKET_CONNECTION_ASSOCIATION_ESTABLISHMENT_MESSAGE = "Connection established with client. Client websocket session: {}. Corresponding server websocket session: {}";
    private static final String WEBSOCKET_CONNECTION_CLOSE_MESSAGE = "Client ({}) websocket session closed. Initiate server websocket session ({}) closing";
    private static final String WEBSOCKET_MESSAGE_SEND_ERROR = "Cannot send websocket message from client session ({}) to server session ({})";
    private static final String WEBSOCKET_MESSAGE_SEND_INFO = "Sending message from client ({}) to server ({})";
    private static final String TRANSPORT_ERROR = "Transport error occurred on client ({}) websocket session";

    public ClientWebSocketConnectionHandler( final WebSocketHandler delegate ) {
        super( delegate );
    }

    @Override
    public void afterConnectionEstablished( final WebSocketSession clientWebSocketSession ) throws Exception {
        super.afterConnectionEstablished( clientWebSocketSession );
        final WebSocketSession serverWebSocketSession = (WebSocketSession) clientWebSocketSession.getAttributes().get( SERVER_WEBSOCKET_SESSION_KEY );
        final ServerWebSocketConnectionHandler serverWebSocketConnectionHandler = (ServerWebSocketConnectionHandler) clientWebSocketSession.getAttributes().get( SERVER_WEBSOCKET_HANDLER_KEY );
        serverWebSocketConnectionHandler.setClientWebSocketSession( clientWebSocketSession );
        LOGGER.info( WEBSOCKET_CONNECTION_ASSOCIATION_ESTABLISHMENT_MESSAGE, clientWebSocketSession, serverWebSocketSession );
    }

    @Override
    public void handleMessage( final WebSocketSession clientWebSocketSession, final WebSocketMessage<?> message ) {
        final WebSocketSession serverWebSocketSession = (WebSocketSession) clientWebSocketSession.getAttributes().get( SERVER_WEBSOCKET_SESSION_KEY );
        LOGGER.debug( WEBSOCKET_MESSAGE_SEND_INFO, clientWebSocketSession, serverWebSocketSession );
        try {
            serverWebSocketSession.sendMessage( message );
        } catch ( IOException e ) {
            LOGGER.error( WEBSOCKET_MESSAGE_SEND_ERROR, serverWebSocketSession, serverWebSocketSession, e );
        }
    }

    @Override
    public void handleTransportError( final WebSocketSession clientWebSocketSession, final Throwable exception ) throws Exception {
        super.handleTransportError( clientWebSocketSession, exception );
        LOGGER.error( TRANSPORT_ERROR, clientWebSocketSession, exception );
        clientWebSocketSession.close();
    }

    @Override
    public void afterConnectionClosed( final WebSocketSession clientWebSocketSession, final CloseStatus closeStatus ) throws Exception {
        super.afterConnectionClosed( clientWebSocketSession, closeStatus );
        final WebSocketSession serverWebSocketSession = (WebSocketSession) clientWebSocketSession.getAttributes().get( SERVER_WEBSOCKET_SESSION_KEY );
        LOGGER.info( WEBSOCKET_CONNECTION_CLOSE_MESSAGE, clientWebSocketSession, serverWebSocketSession );
        serverWebSocketSession.close();
        clientWebSocketSession.getAttributes().remove( SERVER_WEBSOCKET_SESSION_KEY );
        clientWebSocketSession.getAttributes().remove( SERVER_WEBSOCKET_HANDLER_KEY );
    }

    @Override
    public boolean supportsPartialMessages() {
        return true;
    }

}
