package com.acme.wsapigatewaytomcat.websocket;

import java.io.IOException;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.socket.CloseStatus;
import org.springframework.web.socket.TextMessage;
import org.springframework.web.socket.WebSocketMessage;
import org.springframework.web.socket.WebSocketSession;
import org.springframework.web.socket.handler.AbstractWebSocketHandler;

public class ServerWebSocketConnectionHandler extends AbstractWebSocketHandler {

    private static final Logger LOGGER = LoggerFactory.getLogger( ServerWebSocketConnectionHandler.class );
    private static final String DRAINING_QUEUED_MESSAGES_STARTED = "Draining queued messages started";
    private static final String DRAINING_QUEUED_MESSAGES_FINISHED = "Draining queued messages finished";
    private static final String ADDING_MESSAGE_TO_QUEUE = "Adding websocket message to the queue";
    private static final String SERVER_WEBSOCKET_SESSION_KEY = "server-websocket-session-attribute";
    private static final String WEBSOCKET_MESSAGE_SEND_INFO = "Sending message from server ({}) to client ({})";
    private static final String WEBSOCKET_CONNECTION_CLOSE_MESSAGE = "Server ({}) websocket session closed. Initiate client websocket session ({}) closing";
    private static final String CLIENT_SESSION_IS_ALREADY_CLOSED_INFO = "Websocket message cannot be sent from server ({}) to client ({}) as client session is already closed.";
    private static final String WEBSOCKET_MESSAGE_SEND_ERROR = "Cannot send websocket message from server session ({}) to client session ({})";
    private static final String TRANSPORT_ERROR = "Transport error occurred on server ({}) websocket session";

    private final ConcurrentLinkedQueue<WebSocketMessage<?>> messageQueue = new ConcurrentLinkedQueue<>();
    private WebSocketSession clientWebSocketSession;

    @Override
    protected void handleTextMessage( final WebSocketSession serverWebSocketSession, final TextMessage message ) throws Exception {
        if ( isClientWebSocketSessionConnected() ) {
            super.handleTextMessage( serverWebSocketSession, message );
            sendMessageToClient( serverWebSocketSession, message, clientWebSocketSession );
        } else {
            addMessageToQueue( message );
        }
    }

    @Override
    public void handleTransportError( final WebSocketSession serverWebSocketSession, final Throwable exception ) throws Exception {
        super.handleTransportError( serverWebSocketSession, exception );
        LOGGER.error( TRANSPORT_ERROR, serverWebSocketSession, exception );
        serverWebSocketSession.close();
    }

    @Override
    public void afterConnectionClosed( final WebSocketSession serverWebSocketSession, final CloseStatus status ) throws Exception {
        super.afterConnectionClosed( serverWebSocketSession, status );
        if ( isClientWebSocketSessionConnected() ) {
            LOGGER.info( WEBSOCKET_CONNECTION_CLOSE_MESSAGE, serverWebSocketSession, clientWebSocketSession );
            clientWebSocketSession.close();
        } else {
            LOGGER.warn( "After closing the server websocket session ({}) there was no associated client websocket session. ", serverWebSocketSession );
        }
    }

    @Override
    public boolean supportsPartialMessages() {
        return true;
    }

    public void setClientWebSocketSession( final WebSocketSession clientWebSocketSession ) throws IOException {
        drainQueuedMessages( clientWebSocketSession );
        this.clientWebSocketSession = clientWebSocketSession;
    }

    private void drainQueuedMessages( final WebSocketSession clientWebSocketSession ) throws IOException {
        LOGGER.debug( DRAINING_QUEUED_MESSAGES_STARTED );
        WebSocketSession backEndSession = (WebSocketSession) clientWebSocketSession.getAttributes().get( SERVER_WEBSOCKET_SESSION_KEY );
        while ( !messageQueue.isEmpty() ) {
            sendMessageToClient( backEndSession, messageQueue.poll(), clientWebSocketSession );
        }
        LOGGER.debug( DRAINING_QUEUED_MESSAGES_FINISHED );
    }

    private void sendMessageToClient( final WebSocketSession serverWebSocketSession, final WebSocketMessage<?> message, final WebSocketSession clientWebSocketSession ) throws IOException {
        LOGGER.info( WEBSOCKET_MESSAGE_SEND_INFO, serverWebSocketSession, clientWebSocketSession );
        try {
            if ( clientWebSocketSession.isOpen() ) {
                clientWebSocketSession.sendMessage( message );
            } else {
                LOGGER.debug( CLIENT_SESSION_IS_ALREADY_CLOSED_INFO, serverWebSocketSession, clientWebSocketSession );
                serverWebSocketSession.close();
            }
        } catch ( Exception e ) {
            LOGGER.error( WEBSOCKET_MESSAGE_SEND_ERROR, serverWebSocketSession, clientWebSocketSession, e );
            serverWebSocketSession.close();
        }
    }

    private boolean isClientWebSocketSessionConnected() {
        return clientWebSocketSession != null;
    }

    private void addMessageToQueue( final WebSocketMessage<?> webSocketMessage ) {
        LOGGER.debug( ADDING_MESSAGE_TO_QUEUE );
        messageQueue.offer( webSocketMessage );
    }

}
