package com.acme.wsapigatewaytomcat.websocket;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.server.ServerHttpRequest;
import org.springframework.http.server.ServerHttpResponse;
import org.springframework.http.server.ServletServerHttpRequest;
import org.springframework.http.server.ServletServerHttpResponse;
import org.springframework.util.AntPathMatcher;
import org.springframework.util.Assert;
import org.springframework.util.PathMatcher;
import org.springframework.web.socket.WebSocketExtension;
import org.springframework.web.socket.WebSocketHandler;
import org.springframework.web.socket.WebSocketHttpHeaders;
import org.springframework.web.socket.WebSocketSession;
import org.springframework.web.socket.client.WebSocketClient;
import org.springframework.web.socket.server.HandshakeInterceptor;
import org.springframework.web.util.UriComponentsBuilder;

import com.github.mthizo247.cloud.netflix.zuul.web.proxytarget.ProxyTargetResolver;
import com.github.mthizo247.cloud.netflix.zuul.web.socket.ZuulWebSocketProperties;

public class WebsocketHandshakeInterceptor implements HandshakeInterceptor {

    private static final Logger LOGGER = LoggerFactory.getLogger( WebsocketHandshakeInterceptor.class );
    private static final String WS_BROKERAGE_MUST_NOT_BE_NULL = "wsBrokerage must not be null";
    private static final String ROUTE_TARGET_MUST_NOT_BE_NULL = "routeTarget must not be null";
    private static final String CLIENT_WEBSOCKET_HANDSHAKE_FAILURE_ERROR_MESSAGE = "Websocket handshake failed  between the client and the API gateway. Server websocket session is going to be closed. Status code: {}";
    private static final String SERVLET_SERVER_HTTP_REQUEST_REQUIRED = "ServletServerHttpRequest required";
    private static final String WEB_SOCKET_EXTENSIONS_HEADER_WRAPPER_REQUIRED = "WebSocketExtensionsHeaderWrapper required";
    private static final String SEC_WEBSOCKET_EXTENSIONS = "sec-websocket-extensions";
    private static final String SERVER_WEBSOCKET_SESSION_ATTRIBUTE_KEY = "server-websocket-session-attribute";
    private static final String SERVER_WEBSOCKET_SESSION_KEY = "server-websocket-session";
    private static final String SERVER_WEBSOCKET_HANDLER_KEY = "server-websocket-handler";

    private final ProxyTargetResolver proxyTargetResolver;
    private final ZuulWebSocketProperties zuulWebSocketProperties;
    private final WebSocketClient webSocketClient;

    public WebsocketHandshakeInterceptor( final ProxyTargetResolver proxyTargetResolver, final ZuulWebSocketProperties zuulWebSocketProperties, final WebSocketClient webSocketClient ) {
        this.proxyTargetResolver = proxyTargetResolver;
        this.zuulWebSocketProperties = zuulWebSocketProperties;
        this.webSocketClient = webSocketClient;
    }

    public boolean beforeHandshake( final ServerHttpRequest serverHttpRequest, final ServerHttpResponse serverHttpResponse, final WebSocketHandler webSocketHandler, final Map<String, Object> map ) throws Exception {
        LOGGER.info( "Initiating websocket connection to the targeted server" );
        try {
            establishServerWebSocketConnection( serverHttpRequest, map );
        } catch ( Exception e ) {
            LOGGER.error( "Unexpected error during establishing websocket connection to the targeted server: ", e );
            return false;
        }
        return true;
    }

    public void afterHandshake( final ServerHttpRequest serverHttpRequest, final ServerHttpResponse serverHttpResponse, final WebSocketHandler webSocketHandler, final Exception e ) {
        LOGGER.info( "Verifying websocket connection to the targeted server" );
        final int status = ((ServletServerHttpResponse) serverHttpResponse).getServletResponse().getStatus();
        if ( status != 101 ) {
            LOGGER.warn( CLIENT_WEBSOCKET_HANDSHAKE_FAILURE_ERROR_MESSAGE, status );
            final WebSocketSession serverWebSocketSession = (WebSocketSession) getHttpServletRequest( serverHttpRequest ).getAttribute( SERVER_WEBSOCKET_SESSION_ATTRIBUTE_KEY );
            try {
                if ( serverWebSocketSession != null ) {
                    serverWebSocketSession.close();
                }
            } catch ( IOException ex ) {
                LOGGER.error( "Unable to close server websocket session", ex );
            }
        }
    }

    private void establishServerWebSocketConnection( final ServerHttpRequest serverHttpRequest, final Map<String, Object> map ) throws Exception {
        final URI uri = createProxyURI( serverHttpRequest );
        final WebSocketHttpHeaders webSocketHttpHeaders = new WebSocketHttpHeaders();
        final ServerWebSocketConnectionHandler serverWebSocketConnectionHandler = new ServerWebSocketConnectionHandler();
        final WebSocketSession serverWebSocketSession = webSocketClient.doHandshake( serverWebSocketConnectionHandler, webSocketHttpHeaders, uri ).get();
        saveServerEndSessionToHttpRequest( serverHttpRequest, serverWebSocketSession );
        setExtensionsHeaderForOriginalRequest( serverHttpRequest, serverWebSocketSession );
        connectServerSessionToClientSession( map, serverWebSocketConnectionHandler, serverWebSocketSession );
    }

    private URI createProxyURI( final ServerHttpRequest serverHttpRequest ) throws Exception {
        final URI uri = serverHttpRequest.getURI();
        final ZuulWebSocketProperties.WsBrokerage wsBrokerage = getWebSocketBrokerage( uri );
        Assert.notNull( wsBrokerage, WS_BROKERAGE_MUST_NOT_BE_NULL );
        final URI routeTarget = proxyTargetResolver.resolveTarget( wsBrokerage );
        Assert.notNull( routeTarget, ROUTE_TARGET_MUST_NOT_BE_NULL );
        final UriComponentsBuilder builder = UriComponentsBuilder.newInstance()
                .scheme( "ws" )
                .host( routeTarget.getHost() )
                .port( routeTarget.getPort() )
                .path( uri.getPath() );
        if ( StringUtils.isNotEmpty( uri.getQuery() ) ) {
            builder.queryParam( uri.getQuery() );
        }
        final String url = builder.build().toString();
        return new URI( url );
    }

    private ZuulWebSocketProperties.WsBrokerage getWebSocketBrokerage( final URI uri ) {
        String path = uri.toString();
        if ( path.contains( ":" ) ) {
            path = UriComponentsBuilder.fromUriString( path ).build().getPath();
        }
        for ( Map.Entry<String, ZuulWebSocketProperties.WsBrokerage> entry : zuulWebSocketProperties
                .getBrokerages().entrySet() ) {
            final ZuulWebSocketProperties.WsBrokerage wsBrokerage = entry.getValue();
            if ( wsBrokerage.isEnabled() ) {
                for ( String endPoint : wsBrokerage.getEndPoints() ) {
                    final PathMatcher pathMatcher = new AntPathMatcher();
                    if ( pathMatcher.match( toPattern( endPoint ), path + "/" ) ) {
                        return wsBrokerage;
                    }
                }
            }
        }
        return null;
    }

    private String toPattern( String path ) {
        if ( !path.startsWith( "/" ) ) {
            path = "/" + path;
        }
        return path.endsWith( "/" ) ? (path + "**/") : (path + "/**/");
    }

    private void saveServerEndSessionToHttpRequest( final ServerHttpRequest serverHttpRequest, final WebSocketSession serverWebSocketSession ) {
        getHttpServletRequest( serverHttpRequest ).setAttribute( SERVER_WEBSOCKET_SESSION_ATTRIBUTE_KEY, serverWebSocketSession );
    }

    private void setExtensionsHeaderForOriginalRequest( final ServerHttpRequest serverHttpRequest, WebSocketSession serverSession ) {
        List<WebSocketExtension> webSocketExtensions = serverSession.getExtensions();
        List<String> extensionList = new ArrayList<>();
        for ( WebSocketExtension extension : webSocketExtensions ) {
            extensionList.add( extension.getName() );
        }
        serverHttpRequest.getHeaders().put( SEC_WEBSOCKET_EXTENSIONS, extensionList );
        WebSocketExtensionsHeaderWrapper servletRequest = getHttpServletRequest( serverHttpRequest );
        servletRequest.setExtensions( extensionList );
    }

    private WebSocketExtensionsHeaderWrapper getHttpServletRequest( final ServerHttpRequest request ) {
        Assert.isInstanceOf( ServletServerHttpRequest.class, request, SERVLET_SERVER_HTTP_REQUEST_REQUIRED );
        HttpServletRequest servletRequest = ((ServletServerHttpRequest) request).getServletRequest();
        Assert.isInstanceOf( WebSocketExtensionsHeaderWrapper.class, servletRequest, WEB_SOCKET_EXTENSIONS_HEADER_WRAPPER_REQUIRED );
        return (WebSocketExtensionsHeaderWrapper) servletRequest;
    }

    private void connectServerSessionToClientSession( final Map<String, Object> map, final ServerWebSocketConnectionHandler serverWebSocketConnectionHandler, final WebSocketSession serverWebSocketSession ) {
        map.put( SERVER_WEBSOCKET_SESSION_KEY, serverWebSocketSession );
        map.put( SERVER_WEBSOCKET_HANDLER_KEY, serverWebSocketConnectionHandler );
    }

}
